# PDM
import hypothesis.strategies as st
from hypothesis import given, assume

# LOCAL
from sonatoki.Filters import (
    NimiPu,
    Syllabic,
    Alphabetic,
    NimiKuLili,
    NimiKuSuli,
    Phonotactic,
    NimiLinkuCore,
    NimiPuSynonyms,
    NimiLinkuCommon,
    NimiLinkuObscure,
    NimiLinkuSandbox,
    NimiLinkuUncommon,
)
from sonatoki.Cleaners import Lowercase, ConsecutiveDuplicates
from sonatoki.constants import (
    NIMI_PU_SYNONYMS,
    FALSE_POS_SYLLABIC,
    FALSE_POS_ALPHABETIC,
    words_by_tag,
    words_by_usage,
)


@given(st.sampled_from(list(words_by_tag("book", "pu") | NIMI_PU_SYNONYMS)))
def test_pu_filters_non_overlap(s: str):
    res_pu = NimiPu.filter(s)
    res_synonyms = NimiPuSynonyms.filter(s)
    assert (res_pu + res_synonyms) == 1


@given(
    st.sampled_from(
        list(words_by_tag("book", "ku suli") | words_by_tag("book", "ku lili"))
    )
)
def test_ku_filters_non_overlap(s: str):
    s = Lowercase.clean(s)
    s = ConsecutiveDuplicates.clean(s)
    res_ku_suli = NimiKuSuli.filter(s)
    res_ku_lili = NimiKuLili.filter(s)
    assert (res_ku_suli + res_ku_lili) == 1


@given(st.sampled_from(list(words_by_usage(0))))
def test_linku_filters_non_overlap(s: str):
    _ = assume(s != "su")

    s = Lowercase.clean(s)
    s = ConsecutiveDuplicates.clean(s)

    res_core = NimiLinkuCore.filter(s)
    res_common = NimiLinkuCommon.filter(s)
    res_uncommon = NimiLinkuUncommon.filter(s)
    res_obscure = NimiLinkuObscure.filter(s)
    res_sandbox = NimiLinkuSandbox.filter(s)

    assert (res_core + res_common + res_uncommon + res_obscure + res_sandbox) == 1


@given(st.sampled_from(list(words_by_usage(30))))
def test_nimi_linku_properties(s: str):
    assert ConsecutiveDuplicates.clean(s) == s, repr(s)
    assert Alphabetic.filter(s), repr(s)
    assert Syllabic.filter(s), repr(s)
    assert Phonotactic.filter(s), repr(s)
    # Passing phonotactic implies all of the above


@given(st.sampled_from(list(FALSE_POS_ALPHABETIC)))
def test_false_pos_properties(s: str):
    res_syllabic = Syllabic.filter(s)
    res_alphabetic = Alphabetic.filter(s)
    assert res_alphabetic and not res_syllabic



#!/bin/env python3
# STL
import os
import json
import argparse
from typing import Any, Set, Dict, List

# PDM
import emoji
import requests

# LOCAL
from sonatoki.utils import find_unicode_ranges
from sonatoki.Filters import (
    Or,
    LongSyllabic,
    NimiLinkuCore,
    LongAlphabetic,
    NimiLinkuCommon,
    NimiLinkuObscure,
    NimiLinkuUncommon,
)
from sonatoki.Cleaners import ConsecutiveDuplicates
from sonatoki.constants import (
    UCSUR_PUNCT_RANGES,
    UNICODE_PUNCT_RANGES,
    EMOJI_VARIATION_SELECTOR_RANGES,
)

HERE = os.path.dirname(os.path.realpath(__file__))

UNICODE_DATA = "https://unicode.org/Public/UNIDATA/UnicodeData.txt"

LINKU_WORDS = "https://api.linku.la/v1/words?lang=en"
LINKU_SANDBOX = "https://api.linku.la/v1/sandbox?lang=en"

WORDS_10K = "https://raw.githubusercontent.com/first20hours/google-10000-english/master/google-10000-english.txt"
WORDS_25K = "https://raw.githubusercontent.com/dolph/dictionary/master/popular.txt"
WORDS_479K = (
    "https://raw.githubusercontent.com/dwyl/english-words/master/words_alpha.txt"
)

HEADERS = {  # pretend to be Chrome 121, just in case
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/120.0.0.0 Safari/537.3"
}


def download(url: str) -> str:
    if not url.startswith("https://"):
        raise ValueError(url)

    resp = requests.get(url, timeout=5, headers=HEADERS)
    return resp.text


def download_json(url: str) -> Dict[str, Any]:
    resp = download(url)
    return json.loads(resp)


def regen_linku_data():
    data = download_json(LINKU_WORDS)
    with open(os.path.join(HERE, "linku.json"), "w") as f:
        _ = f.write(json.dumps(data))

    data = download_json(LINKU_SANDBOX)
    with open(os.path.join(HERE, "sandbox.json"), "w") as f:
        _ = f.write(json.dumps(data))


def regen_false_negatives():
    # TODO: regen from my frequency data where the score is below 0.8?
    KnownWords = Or(
        NimiLinkuCore,
        NimiLinkuCommon,
        NimiLinkuUncommon,
        NimiLinkuObscure,
    )

    syllabic_matches: Set[str] = set()
    alphabetic_matches: Set[str] = set()
    data = download(WORDS_25K)
    for word in data.splitlines():
        if not word:
            continue
        word = ConsecutiveDuplicates.clean(word)

        if KnownWords.filter(word):
            # ignore dictionary
            continue
        if LongSyllabic.filter(word):
            syllabic_matches.add(word)
            continue
        if LongAlphabetic.filter(word):
            alphabetic_matches.add(word)
            continue

    # TODO: include short matches or no?
    with open(os.path.join(HERE, "syllabic.txt"), "w") as f:
        syllabic_final = sorted([word + "\n" for word in syllabic_matches])
        f.writelines(syllabic_final)

    with open(os.path.join(HERE, "alphabetic.txt"), "w") as f:
        alphabetic_final = sorted([word + "\n" for word in alphabetic_matches])
        f.writelines(alphabetic_final)


def regen_unicode_data():
    PUNCT_CATEGORIES = {
        # Punctuation
        "Pc",  # Connector
        "Pd",  # Dash
        "Pe",  # Close (end)
        "Pf",  # Final
        "Pi",  # Initial
        "Po",  # Other
        "Ps",  # Open (sOpen)
        # Symbol
        "Sm",  # Math
        "Sk",  # Modifier (kModifier)
        "Sc",  # Currency
        "So",  # Other
    }
    r"""These characters are in Symbol other (So) but are not in
    `\p{Punctuation}` However, I began excluding them again, because it turns
    out that some sequences of latin alphabet emoji."""

    # NOTE: There are many characters which look like writing characters but are in the punctuation character class. Examples:
    # - kangxi radicals from ⺀ to ⿕ which are for demonstration, not writing
    # - parenthesized hangul letters and syllables from ㈀ to ㈜
    # - circled katakana from ㋐ to ㋾
    # the latter two shouldn't be in `\p{Punctuation}` if the latin alphabet isn't... oof

    def is_punctuation(data: List[str]):
        return data[2] in PUNCT_CATEGORIES

    def get_character(data: List[str]):
        return chr(int(data[0], 16))

    unicode_data = download(UNICODE_DATA)
    unicode_punctuation = ""
    for line in unicode_data.split("\n"):
        if not line:  # damn you, trailing newline
            continue
        # NOTE: UnicodeData.txt lists a range if there are many consecutive similar characters
        # (e.g. CJK Ideograph, First at 4E00 and CJK Ideograph, Last at 9FFF).
        # This does not apply to any currently defined punctuation category.

        unicode_data = line.split(";")
        if not is_punctuation(unicode_data):
            continue

        char = get_character(unicode_data)

        unicode_punctuation += char

    unicode_punctuation = emoji.replace_emoji(unicode_punctuation)

    unicode_ranges = find_unicode_ranges(unicode_punctuation)
    unicode_ranges.extend(UCSUR_PUNCT_RANGES)
    # unicode_ranges.extend(EMOJI_VARIATION_SELECTOR_RANGES)  # made unnecessary by emoji library
    unicode_ranges = sorted(unicode_ranges)
    # sorted in case my manual additions are out of order

    if unicode_ranges != UNICODE_PUNCT_RANGES:
        output = json.dumps(unicode_ranges, indent=4, ensure_ascii=True)
        print(output)


def main(argv: argparse.Namespace):
    regen_unicode_data()
    regen_linku_data()
    regen_false_negatives()


if __name__ == "__main__":
    """Helper script to fetch UNICODE_PUNCT in constants.py."""
    parser = argparse.ArgumentParser()

    # TODO: choice between regen unicode data, regen linku, regen english phonomatches
    argv = parser.parse_args()
    main(argv)



# STL
import string

# PDM
import hypothesis.strategies as st
from hypothesis import given, example

# LOCAL
from sonatoki.Filters import (
    Or,
    And,
    Not,
    NimiPu,
    PuName,
    Numeric,
    Syllabic,
    Alphabetic,
    NimiKuLili,
    NimiKuSuli,
    Phonotactic,
    Punctuation,
    AlphabeticRe,
    LongSyllabic,
    MemberFilter,
    NimiLinkuCore,
    PunctuationRe,
    LongAlphabetic,
    PunctuationRe1,
    LongPhonotactic,
    NimiLinkuCommon,
    FalsePosSyllabic,
    NimiLinkuObscure,
    NimiLinkuSandbox,
    NimiLinkuUncommon,
)
from sonatoki.Cleaners import Lowercase, ConsecutiveDuplicates
from sonatoki.constants import FALSE_POS_SYLLABIC, words_by_tag

# FILESYSTEM
from .test_utils import PROPER_NAME_RE


@given(st.sampled_from(list(words_by_tag("book", "pu"))))
@example("lukin")
@example("selo")
@example("li")
def test_NimiPu(s: str):
    res = NimiPu.filter(s)
    assert res, repr(s)


@given(st.sampled_from(list(words_by_tag("usage_category", "core"))))
@example("pona")
def test_NimiLinkuCore(s: str):
    res = NimiLinkuCore.filter(s)
    assert res, repr(s)


@given(st.sampled_from(list(words_by_tag("usage_category", "common"))))
@example("n")
@example("tonsi")
@example("kipisi")
def test_NimiLinkuCommon(s: str):
    res = NimiLinkuCommon.filter(s)
    assert res, repr(s)


@given(st.sampled_from(list(words_by_tag("usage_category", "uncommon"))))
def test_NimiLinkuUncommon(s: str):
    res = NimiLinkuUncommon.filter(s)
    assert res, repr(s)


@given(st.sampled_from(list(words_by_tag("usage_category", "obscure"))))
@example("pake")
@example("san")
def test_NimiLinkuObscure(s: str):
    res = NimiLinkuObscure.filter(s)
    assert res, repr(s)


@given(st.sampled_from(list(words_by_tag("usage_category", "sandbox"))))
@example("kalamARR")
@example("Pingo")
def test_NimiLinkuSandbox(s: str):
    s = Lowercase.clean(s)
    s = ConsecutiveDuplicates.clean(s)
    # above two are necessary due to kalamARR and Pingo
    res = NimiLinkuSandbox.filter(s)
    assert res, repr(s)


@given(st.from_regex(Phonotactic.pattern, fullmatch=True))
@example("kijetesantakalu")
@example("n")
def test_Phonotactic(s: str):
    res = Phonotactic.filter(s)
    assert res, repr(s)


@given(st.from_regex(Phonotactic.pattern, fullmatch=True))
def test_LongPhonotactic(s: str):
    len_ok = len(s) >= LongPhonotactic.length
    res = LongPhonotactic.filter(s)
    assert res == len_ok, repr(s)  # will match given fullmatch


@given(st.from_regex(Syllabic.pattern, fullmatch=True))
@example("wuwojitiwunwonjintinmanna")
def test_Syllabic(s: str):
    res = Syllabic.filter(s)
    assert res, repr(s)


@given(st.from_regex(Syllabic.pattern, fullmatch=True))
def test_LongSyllabic(s: str):
    len_ok = len(s) >= LongSyllabic.length
    res = LongSyllabic.filter(s)
    assert res == len_ok


@given(st.from_regex(AlphabeticRe.pattern, fullmatch=True))
@example("muems")
@example("mpptp")
@example("tptpt")
def test_Alphabetic(s: str):
    res_fn = Alphabetic.filter(s)
    res_re = AlphabeticRe.filter(s)
    assert res_fn == res_re, repr(s)


@given(st.from_regex(AlphabeticRe.pattern, fullmatch=True))
def test_LongAlphabetic(s: str):
    len_ok = len(s) >= LongAlphabetic.length
    res = LongAlphabetic.filter(s)
    assert res == len_ok


@given(st.from_regex(AlphabeticRe.pattern, fullmatch=True))
def test_AlphabeticRe(s: str):
    res_re = AlphabeticRe.filter(s)
    assert res_re, repr(s)


@given(st.from_regex(PROPER_NAME_RE, fullmatch=True))
def test_ProperName(s: str):
    res = PuName.filter(s)
    assert res, repr(s)


@given(st.from_regex(PunctuationRe.pattern, fullmatch=True))
@example("[]")
@example(r"\\")
@example(r"\"")
@example("⟨·⟩")
@example("…")
@example("「」")
@example(string.punctuation)
def test_PunctuationRe1(s: str):
    res = PunctuationRe1.filter(s)
    assert res, repr(s)


@given(st.from_regex(PunctuationRe.pattern, fullmatch=True))
def test_PunctuationRe(s: str):
    res_re = PunctuationRe.filter(s)
    res_re1 = PunctuationRe1.filter(s)
    assert res_re == res_re1, repr(s)


@given(st.from_regex(PunctuationRe.pattern, fullmatch=True))
@example("\U000f1990")  # UCSUR char
def test_Punctuation(s: str):
    res_fn = Punctuation.filter(s)
    res_re1 = PunctuationRe1.filter(s)
    assert res_fn == res_re1, repr(s)


@given(st.from_regex(r"\d+", fullmatch=True))
@example("124125")
@example("99990000")
def test_Numeric(s: str):
    res = Numeric.filter(s)
    assert res, repr(s)


@given(
    st.from_regex(PunctuationRe.pattern, fullmatch=True)
    | st.from_regex(r"\d+", fullmatch=True),
)
def test_OrFilter(s: str):
    filter = Or(Punctuation, Numeric)
    res = filter.filter(s)
    res_punctuation = Punctuation.filter(s)
    res_numeric = Numeric.filter(s)
    assert res and (res_punctuation or res_numeric)


# NOTE: No subset filter test because A | B is not the same as A combined with B.
# e.g. "apple" passes Alphabetic, "..." passes Punctuation, "apple..." passes neither
# but would incorrectly pass a combined filter.
@given(
    st.sampled_from(
        list(words_by_tag("book", "pu") | words_by_tag("usage_category", "obscure"))
    )
)
def test_MemberFilters_OrFilter(s: str):
    filter = Or(NimiPu, NimiLinkuObscure)
    assert issubclass(filter, MemberFilter)

    res = filter.filter(s)
    res_pu = NimiPu.filter(s)
    res_obscure = NimiLinkuObscure.filter(s)
    assert res and (res_pu or res_obscure)


@given(
    st.sampled_from(
        list(
            words_by_tag("book", "ku suli")
            | words_by_tag("book", "ku lili")
            | words_by_tag("usage_category", "uncommon")
            | words_by_tag("usage_category", "obscure")
            | words_by_tag("usage_category", "sandbox")
        ),
    )
)
def test_OrFilter_IsipinEpiku(s: str):
    filter = Or(
        NimiKuSuli, NimiKuLili, NimiLinkuUncommon, NimiLinkuObscure, NimiLinkuSandbox
    )

    s = Lowercase.clean(s)
    s = ConsecutiveDuplicates.clean(s)

    res = filter.filter(s)
    res_ku_suli = NimiKuSuli.filter(s)
    res_ku_lili = NimiKuLili.filter(s)
    res_uncommon = NimiLinkuUncommon.filter(s)
    res_obscure = NimiLinkuObscure.filter(s)
    res_sandbox = NimiLinkuSandbox.filter(s)
    assert res and (
        res_ku_suli or res_ku_lili or res_uncommon or res_obscure or res_sandbox
    )


@given(st.sampled_from(list(words_by_tag("book", "pu"))))
def test_AndFilter(s: str):
    s = s.capitalize()
    f = And(PuName, NimiPu)
    assert f.filter(s)


@given(st.sampled_from(list(words_by_tag("book", "pu"))))
def test_NotFilter(s: str):
    f = Not(NimiPu)
    assert not f.filter(s)


@given(
    st.sampled_from(list(FALSE_POS_SYLLABIC))
    | st.from_regex(Syllabic.pattern, fullmatch=True)
    | st.from_regex(AlphabeticRe.pattern, fullmatch=True)
)
def test_AndNotFilter(s: str):
    AndNotFilter = And(Syllabic, Not(FalsePosSyllabic))

    res_syl = Syllabic.filter(s)
    res_fp = FalsePosSyllabic.filter(s)
    res_composed = AndNotFilter.filter(s)

    if not res_syl:
        # if it isn't syllabic in the first place, it shouldn't match anything
        assert not res_fp and not res_syl and not res_composed

    if res_fp:
        # syl matched- but if fp matches, then the composed filter should not match
        assert not res_composed


@given(
    st.sampled_from(list(words_by_tag("book", "pu") | words_by_tag("book", "ku suli")))
)
def test_AddTokensToMemberFilter(s: str):
    PuEnKuSuliFilter = NimiPu(add=NimiKuSuli.tokens)
    assert PuEnKuSuliFilter.filter(s)


@given(
    st.sampled_from(
        list(
            words_by_tag("usage_category", "sandbox") | words_by_tag("book", "ku lili")
        )
    )
)
def test_AddTokensToMemberFilterNegative(s: str):
    PuEnKuSuliFilter = NimiPu(add=NimiKuSuli.tokens)
    assert not PuEnKuSuliFilter.filter(s)


@given(
    st.sampled_from(
        list(
            words_by_tag("book", "pu")
            | words_by_tag("book", "ku suli")
            | words_by_tag("book", "ku lili")
            | words_by_tag("usage_category", "uncommon")
            | words_by_tag("usage_category", "obscure")
            | words_by_tag("usage_category", "sandbox")
        ),
    )
    | st.from_regex(Syllabic.pattern, fullmatch=True)
)
def test_SubTokensFromMemberFilter(s: str):
    NimiAlaFilter = NimiLinkuCore(sub=NimiPu.tokens)
    # core is a strict subset of pu
    # if kin becomes core, needs to be corrected

    assert not NimiAlaFilter.filter(s)



# STL
from typing import List, Tuple

# PDM
import pytest

# LOCAL
from sonatoki.ilo import Ilo
from sonatoki.Configs import LazyConfig, PrefConfig, CorpusConfig


@pytest.fixture
def ilo() -> Ilo:
    ilo = Ilo(**PrefConfig)
    return ilo


@pytest.fixture()
def lazy_ilo() -> Ilo:
    ilo = Ilo(**LazyConfig)
    return ilo


@pytest.fixture()
def corpus_ilo() -> Ilo:
    ilo = Ilo(**CorpusConfig)
    return ilo


ALL_VALID = [
    "mi unpa e mama sina",
    "mama sina li lon seme? mi wile toki tawa ona",
    "sina sike pakala",
    "    sina    seme     e     mi     ?",
    "AAAAAAAAAAA",
    "muuuu MUUU muUuUuU",
    "wawa mute. " * 10,
    "󱥄󱥬󱥩󱤴",  # "o toki tawa mi" in UCSUR
    "󱤴󱤧󱤑󱥍󱦗󱤖󱥡󱦘󱤬󱥭‍󱥡󱥚",
    "󱤑󱦐󱥗󱦜󱦈󱦜󱥉󱦜󱦑󱥄󱤤󱤂󱤉󱥆󱤀",
    "o lukin, 󱤴󱥬󱥩󱤴󱤧wawa",
    "ni li sona kiwen",
    "nimi namako li toki e ale",
    "mi open mute a",  # mostly eng words
]

IGNORABLES = [
    "",
    " ",
    "2+2=5",
    "kiwen moli 42",
    "https://mun.la/sona",
    "https://example.com/",
    "mi wile e ni: <https://example.com> li pona",
    "lipu https://example.com li kama pona",
    "<:owe:843315277286473778><:owe:843315277286473778><:owe:843315277286473778><:owe:843315277286473778><:owe:843315277286473778>",
    "...",
    " ⟨·⟩, a",
    "·····",
    "o lukin: [[w:QWERTY]]",
    "nasa la mi ken pana e ni: <unrelated_words_that_are_illegal>",
    "❤️",  # heart
    "😊",
    "👨‍👩‍👧‍👧",  # family emoji with zwj
    # every non-emoji in the writables
    "🄀🄁🄂🄃🄄🄅🄆🄇🄈🄉🄊🄋🄌🄍🄎🄏🄐🄑🄒🄓🄔🄕🄖🄗🄘🄙🄚🄛🄜🄝🄞🄟🄠🄡🄢🄣🄤🄥🄦🄧🄨🄩🄪🄫🄬🄭🄮🄯🄰🄱🄲🄳🄴🄵🄶🄷🄸🄹🄺🄻🄼🄽🄾🄿🅀🅁🅂🅃🅄🅅🅆🅇🅈🅉🅊🅋🅌🅍🅎🅏🅐🅑🅒🅓🅔🅕🅖🅗🅘🅙🅚🅛🅜🅝🅞🅟🅠🅡🅢🅣🅤🅥🅦🅧🅨🅩🅪🅫🅬🅭🅮🅯🅲🅳🅴🅵🅶🅷🅸🅹🅺🅻🅼🅽🆀🆁🆂🆃🆄🆅🆆🆇🆈🆉🆊🆋🆌🆍🆏🆐 🆛🆜🆝🆞🆟🆠🆡🆢🆣🆤🆥🆦🆧🆨🆩🆪🆫🆬🆭🇦🇧🇨🇩🇪🇫🇬🇭🇮🇯🇰🇱🇲🇳🇴🇵🇶🇷🇸🇹🇺🇻🇼🇽🇾🇿",
    "🅰️🅱️🅾️🅱️🅰️",  # blood type emojis
]

SYLLABIC_MATCHES = [
    "ni li tenpo penpo",
    "sipisi",
    "walawa malama walama malama mupi",
    "mi sona ala e nimi sunopatikuna",
    "kalama wuwojiti li pana e sona",
    "jan Awaja en jan Alasali en jan Akesinu li pona",  # syllables match before names here
    "jan Ke Tami",
    "kulupu Kuko",
]

ALPHABETIC_MATCHES = [
    "mi mtue o kama sona",
    "mi mute o kma son",  # this one is odd because `son` is an unintended phonetic match
    "mi mute o kama kne snoa a",
    "ni li tptpt",
    "mi wile pana lon sptp",
    "tmo tawa mi li pona mute la mi kepeken ona lon tenpo mute",
    "mi pakla lon nimi pi mute lili, taso ale li pona tan ni: mi toki mute",
]

NAME_MATCHES = [
    "musi Homestuck li ike tawa mi",
    "ilo Google li sona ala e nimi Emoticon la mi wile utala e ona",
    "toki Kanse li lon",
    "toki Lojban li nasa e lawa mi",
    "ilo Firefox",
    "ilo FaceBook li nasa",
    "mi kepeken ilo MySQL",
    "poki li nasin SQLite",
    "mi musi Space Station 13",
    "jan Tepo en jan Salo en jan Lakuse en pipi Kewapi en soweli Eweke en mi li musi",
]

SOME_INVALID = [
    "󱤎󱥁󱤡󱥀󱤬󱥚󱥁󱤧󱤬󱦜󱤬󱥆󱤡󱥞󱤘󱤆󱤉󱥬󱥠󱥞󱦜󱤬󱥓「Input mode」󱤿󱥮󱤧󱤬󱦜󱤘󱤡󱥫󱥁󱤡󱥞󱤙󱤿「Direct」󱦜󱤿󱥁󱤧󱥈",
    "kulupu xerox li ike",
    "mi tawa ma ohio",
    "sina toki e nimi what pi toki Inli",
    "wawa la o lukin e ni: your mom",
]

CORPUS_SPECIFIC = [
    # "ki le konsi si te isipin epiku le pasila to",
    "ki konsi te isipin epiku pasila to",  # the sandbox has not documented si or le
    'jasima omekapo, ki nimisin "jasima enko nimisin". ki enko alu linluwi Jutu alu epiku ki epiku baba is you. ki likujo "SINtelen pona", ki epiku alu "sitelen pona". ki kepen wawajete isipin, kin ki yupekosi alu lipamanka alu wawajete, kin ki enko isipin lipamanka linluwi alu wawajete',
    "kalamARRRR",
    "Pingo",
    "we Luke li alente wa",
]
CORPUS_SPECIFIC_XFAIL: List[str] = []


EXCESSIVE_SYLLABICS = [
    # NOTE: this is sometimes hard to distinguish from EXCESSIVE_ENGLISH
    "manama manama namana namana majani makala",
    "I manipulate a passe pile so a ton emulate, akin to intake",
    "a ton of insolate puke. make no amen, no joke.",
    "I elope so, to an elite untaken tune, some unwise tone",
    "insane asinine lemon awesome atone joke",
    "insane asinine lemon awesome atone",  # i got more clever
    "nope, no, joke",
    "insane",
    "woman",
    "man",
    "opposite",
    "nine emo women see anime alone",
    "i like mini potato",
]

EXCESSIVE_ALPHABETICS = [
    "wen i tok usin onli notes in toki pona i look silli. ",
    "I wait, I sulk, as a tool I make stoops to ineptness.",
    "aaa i non-saw usa's most multiple element-set. it's as asinine as in `e`-less speak",
    "so, to atone like papa—an awesome anon (no-name) sin man—i ate an asinine lemon-limelike tomato jalapeno isotope. 'nonsense!' amen. note to Oman: take mine katana to imitate a ninja in pantomime. atomise one nuke? 'insane misuse!' same. likewise, Susan, awaken a pepino melon in a linen pipeline. (penile) emanate semen. joke: manipulate a tame toneme to elope online tonite",
]

EXCESSIVE_TYPOES = [
    "mi pakla ln tepo mtue ls mi kn ala tok poan aun seem",
    "sina poan",
]

EXCESSIVE_NAMES = [
    "I Want To Evade The Filter",
    "If You Do This The Bot Can't See You",
    "This Is A Statement In Perfect Toki Pona, I Guarantee",
]

EXCESSIVE_ENGLISH = [
    "me when i tawa sike",  # previous false positive; fixed by english ignorables
    "Maybe I’m too nasa",  # previous false positive; fixed by LongSyllabic and LongAlphabetic
    "I see :)",
    "I wanna see",  # same down to here
    "i'm online all the time",
    "How to Cut a Kiwi",
    "a e i o u",
    "21st",  # previous false positive; fixed by ProperName change
]

NON_MATCHES = [
    "bong",
    "super bruh moment 64",
    "homestuck",
    "homestuck Homestuck",
    "what if i went to the store ",
]

KNOWN_GOOD = (
    ALL_VALID
    + SYLLABIC_MATCHES
    + ALPHABETIC_MATCHES
    + NAME_MATCHES
    + SOME_INVALID
    + IGNORABLES
)

KNOWN_BAD = (
    EXCESSIVE_SYLLABICS
    + EXCESSIVE_ALPHABETICS
    + EXCESSIVE_NAMES
    + EXCESSIVE_TYPOES
    + EXCESSIVE_ENGLISH
    + NON_MATCHES
)

FALSE_NEGATIVES = [
    # emoticon should not be a problem
    # a token that is one edit off a known word should be allowed
    "mi pnoa",
    "tok",
    "mut",
    "poan",
    "mtue",
    "mi nasa B^)",  # emoticon
    "lete li ike x.x",  # this is an emoticon but passes because 'x' is in Filters.Miscellaneous
    "😃⃢👍",  # sincerely, no idea, but it came up and it should be omitted by emojis but isn't
]

FALSE_POSITIVES = [
    "Knowing a little toki pona",  # name, dict, alphabet, dict, dict- damn, that's hard.
]

IGNORABLE_PAIRS: List[Tuple[str, str]] = [
    ("o lukin e ni: https://example.com/", "o lukin e ni:"),
    ("ni li nasa anu seme <:musiwawa:198591138591>", "ni li nasa anu seme"),
    ("seme la ni li toki pona ala https://example.com/", "seme la ni li toki pona ala"),
    ("```\ndef bad():\n    pass\n``` o lukin e ni", "o lukin e ni"),
    ("mi tawa tomo telo 💦💦", "mi tawa tomo telo"),
    ("o lukin e lipu ni: [[wp:Canvassing]]", "o lukin e lipu ni:"),
]


@pytest.mark.parametrize("text", KNOWN_GOOD)
def test_known_good_pref(ilo: Ilo, text: str):
    assert ilo.is_toki_pona(text), text


@pytest.mark.parametrize("text", KNOWN_BAD + CORPUS_SPECIFIC)
def test_known_bad_pref(ilo: Ilo, text: str):
    assert not ilo.is_toki_pona(text), text


@pytest.mark.parametrize("text", KNOWN_GOOD + CORPUS_SPECIFIC)
def test_known_good_corpus(corpus_ilo: Ilo, text: str):
    assert corpus_ilo.is_toki_pona(text), text


@pytest.mark.parametrize("text", KNOWN_BAD)
def test_known_bad_corpus(corpus_ilo: Ilo, text: str):
    assert not corpus_ilo.is_toki_pona(text), text


@pytest.mark.parametrize("text", KNOWN_GOOD)
def test_known_good_lazy(lazy_ilo: Ilo, text: str):
    assert lazy_ilo.is_toki_pona(text), text
    # assumption: lazy ilo should pass anything the more strict ilo does


@pytest.mark.parametrize("text", NON_MATCHES)
def test_known_bad_lazy(lazy_ilo: Ilo, text: str):
    assert not lazy_ilo.is_toki_pona(text), text


# yes this set manip is silly
@pytest.mark.parametrize("text", list(set(KNOWN_BAD) - set(NON_MATCHES)))
def test_weakness_of_lazy(lazy_ilo: Ilo, text: str):
    # NOTE: This is demonstrative, not preferential
    assert lazy_ilo.is_toki_pona(text), text


@pytest.mark.xfail
@pytest.mark.parametrize("text", FALSE_POSITIVES)
def test_false_positives_pref(ilo: Ilo, text: str):
    assert not ilo.is_toki_pona(text)


@pytest.mark.xfail
@pytest.mark.parametrize("text", FALSE_NEGATIVES)
def test_false_negatives_pref(ilo: Ilo, text: str):
    assert ilo.is_toki_pona(text)


@pytest.mark.xfail
@pytest.mark.parametrize("text", CORPUS_SPECIFIC_XFAIL)
def test_false_positives_corpus(corpus_ilo: Ilo, text: str):
    assert not corpus_ilo.is_toki_pona(text)


@pytest.mark.parametrize("pair", IGNORABLE_PAIRS)
def test_pref_ignorable_doesnt_change_score(ilo: Ilo, pair: Tuple[str, str]):
    with_ignorable, without_ignorable = pair
    with_ignorable = ilo.preprocess(with_ignorable)
    without_ignorable = ilo.preprocess(without_ignorable)
    score_with = ilo._is_toki_pona(with_ignorable)["score"]
    score_without = ilo._is_toki_pona(without_ignorable)["score"]
    assert score_with == score_without


@pytest.mark.parametrize("pair", IGNORABLE_PAIRS)
def test_lazy_ignorable_doesnt_change_score(lazy_ilo: Ilo, pair: Tuple[str, str]):
    with_ignorable, without_ignorable = pair
    with_ignorable = lazy_ilo.preprocess(with_ignorable)
    without_ignorable = lazy_ilo.preprocess(without_ignorable)
    score_with = lazy_ilo._is_toki_pona(with_ignorable)["score"]
    score_without = lazy_ilo._is_toki_pona(without_ignorable)["score"]
    assert score_with == score_without


@pytest.mark.parametrize("pair", IGNORABLE_PAIRS)
def test_corpus_ignorable_doesnt_change_score(corpus_ilo: Ilo, pair: Tuple[str, str]):
    with_ignorable, without_ignorable = pair
    with_ignorable = corpus_ilo.preprocess(with_ignorable)
    without_ignorable = corpus_ilo.preprocess(without_ignorable)
    score_with = corpus_ilo._is_toki_pona(with_ignorable)["score"]
    score_without = corpus_ilo._is_toki_pona(without_ignorable)["score"]
    assert score_with == score_without



# STL
from typing import List, TypedDict

# PDM
import yaml
import pytest

# LOCAL
from sonatoki.Tokenizers import (
    SentTokenizer,
    WordTokenizer,
    SentTokenizerRe,
    WordTokenizerRe,
    SentTokenizerRe1,
    WordTokenizerRe1,
)


class TokenizerTest(TypedDict):
    name: str
    input: str
    output: List[str]
    should_be_equal: bool
    xfail: bool


def load_params_from_yaml(json_path: str) -> List[TokenizerTest]:
    with open(json_path) as f:
        return yaml.safe_load(f)


def load_tokenizer_tests(json_path: str) -> List[TokenizerTest]:
    tests = load_params_from_yaml(json_path)
    formatted_tests: List[TokenizerTest] = []
    for test in tests:
        formatted_tests.append(
            TokenizerTest(
                xfail=test.get("xfail", False),
                name=test.get("name", ""),
                input=test["input"],
                output=test.get("output", []),
                should_be_equal=test.get("should_be_equal", True),
            )
        )

    return formatted_tests


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_sentences_tok.yml")
)
def test_SentTokenizer(test: TokenizerTest):
    if test["xfail"]:
        pytest.xfail()

    fn_tokenized = SentTokenizer.tokenize(test["input"])
    re1_tokenized = SentTokenizerRe1.tokenize(test["input"])
    assert fn_tokenized == re1_tokenized, test["name"]


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_sentences_tok.yml")
)
def test_SentTokenizerRe(test: TokenizerTest):
    if test["xfail"]:
        pytest.xfail()

    re_tokenized = SentTokenizerRe.tokenize(test["input"])
    re1_tokenized = SentTokenizerRe1.tokenize(test["input"])
    assert re_tokenized == re1_tokenized, test["name"]


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_sentences_tok.yml")
)
def test_SentTokenizerRe1(test: TokenizerTest):
    if test["xfail"]:
        pytest.xfail()

    re1_tokenized = SentTokenizerRe1.tokenize(test["input"])
    assert re1_tokenized == test["output"], test["name"]


###################
# Word tokenizers #
###################


@pytest.mark.parametrize(
    "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_words_tok.yml")
)
def test_WordTokenizer(test: TokenizerTest):
    if test["xfail"]:
        pytest.xfail()

    fn_tokenized = WordTokenizer.tokenize(test["input"])
    # re1_tokenized = WordTokenizerRe1.tokenize(test["input"])
    # assert fn_tokenized == re1_tokenized, test["name"]
    assert fn_tokenized == test["output"], test["name"]


# @pytest.mark.parametrize(
#     "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_words_tok.yml")
# )
# def test_WordTokenizerRe(test: TokenizerTest):
#     if test["xfail"]:
#         pytest.xfail()
#
#     re_tokenized = WordTokenizerRe.tokenize(test["input"])
#     re1_tokenized = WordTokenizerRe1.tokenize(test["input"])
#     assert re_tokenized == re1_tokenized, test["name"]
#
#
# @pytest.mark.parametrize(
#     "test", load_tokenizer_tests("tests/tokenize_cases/tokenize_words_tok.yml")
# )
# def test_WordTokenizerRe1(test: TokenizerTest):
#     """This implementation will always exhibit the correct behavior, so long as `regex` is up to date
#     Thus, it is used as a reference implementation for all other tests"""
#     if test["xfail"]:
#         pytest.xfail()
#
#     re1_tokenized = WordTokenizerRe1.tokenize(test["input"])
#     assert re1_tokenized == test["output"], test["name"]



# STL
from typing import List, Type

# LOCAL
from sonatoki.types import Number, Scorecard
from sonatoki.Filters import Filter
from sonatoki.Scorers import Scorer, SentNoOp, SentenceScorer
from sonatoki.Cleaners import Cleaner
from sonatoki.Tokenizers import Tokenizer, SentTokenizer, WordTokenizer
from sonatoki.Preprocessors import Preprocessor


class Ilo:
    __preprocessors: List[Type[Preprocessor]]
    __sent_tokenizer: Type[Tokenizer]
    __word_tokenizer: Type[Tokenizer]
    __cleaners: List[Type[Cleaner]]
    __ignoring_filters: List[Type[Filter]]
    __scoring_filters: List[Type[Filter]]
    __scorer: Type[Scorer]
    __sentence_scorer: Type[SentenceScorer]
    __passing_score: Number

    def __init__(
        self,
        preprocessors: List[Type[Preprocessor]],
        cleaners: List[Type[Cleaner]],
        ignoring_filters: List[Type[Filter]],
        scoring_filters: List[Type[Filter]],
        scorer: Type[Scorer],
        passing_score: Number,
        sentence_scorer: Type[SentenceScorer] = SentNoOp,
        word_tokenizer: Type[Tokenizer] = WordTokenizer,
        sent_tokenizer: Type[Tokenizer] = SentTokenizer,
    ):
        super().__init__()
        # avoid keeping a ref to user's list just in case
        self.__preprocessors = [*preprocessors]
        self.__sent_tokenizer = sent_tokenizer
        self.__word_tokenizer = word_tokenizer
        self.__cleaners = [*cleaners]
        self.__ignoring_filters = [*ignoring_filters]
        self.__scoring_filters = [*scoring_filters]
        self.__scorer = scorer
        self.__sentence_scorer = sentence_scorer
        self.__passing_score = passing_score

    def preprocess(self, msg: str) -> str:
        for p in self.__preprocessors:
            msg = p.process(msg)
        return msg

    def word_tokenize(self, msg: str) -> List[str]:
        """It is *highly* recommended that you run `ilo.preprocess` first."""
        return self.__word_tokenizer.tokenize(msg)

    def sent_tokenize(self, msg: str) -> List[str]:
        """It is *highly* recommended that you run `ilo.preprocess` first."""
        return self.__sent_tokenizer.tokenize(msg)

    def clean_token(self, token: str) -> str:
        for c in self.__cleaners:
            token = c.clean(token)
        return token

    def clean_tokens(self, tokens: List[str]) -> List[str]:
        # NOTE: tested, making a new list with a for loop *is* faster than:
        # list comp, generator comp, in-place replacement
        cleaned_tokens: List[str] = list()
        for token in tokens:
            cleaned_token = self.clean_token(token)
            if not cleaned_token:
                # TODO: warn user?
                continue
            cleaned_tokens.append(cleaned_token)
        return cleaned_tokens

    def _filter_token(self, token: str) -> bool:
        for f in self.__ignoring_filters:
            if f.filter(token):
                return True
        return False

    def filter_tokens(self, tokens: List[str]) -> List[str]:
        filtered_tokens: List[str] = []
        for token in tokens:
            if self._filter_token(token):
                continue
            # the ignoring filter is true if the token matches
            # the user wants to ignore these so keep non-matching tokens
            filtered_tokens.append(token)
        return filtered_tokens

    def score_tokens(self, tokens: List[str]) -> float:
        return self.__scorer.score(tokens, self.__scoring_filters)

    def score_sentences(self, scorecards: List[Scorecard]) -> List[Scorecard]:
        return self.__sentence_scorer.score(scorecards)

    def _is_toki_pona(self, message: str) -> Scorecard:
        """Process a message into its tokens, then filters, cleans, and scores
        them. Message must already be preprocessed, normally done in
        `self.is_toki_pona(message)`.

        Returns a `Scorecard` with all changes to the input text and a score.
        """
        tokenized = self.word_tokenize(message)
        filtered = self.filter_tokens(tokenized)
        cleaned = self.clean_tokens(filtered)
        score = self.score_tokens(cleaned)

        scorecard: Scorecard = {
            "text": message,
            "tokenized": tokenized,
            "filtered": filtered,
            "cleaned": cleaned,
            "score": score,
        }

        return scorecard

    def is_toki_pona(self, message: str) -> bool:
        """Determines whether a text is or is not Toki Pona."""
        message = self.preprocess(message)
        scorecard = self._is_toki_pona(message)
        return scorecard["score"] >= self.__passing_score

    def _are_toki_pona(self, message: str) -> List[Scorecard]:
        """Split a message into sentences, then return a list with each
        sentence's scorecard from `self._is_toki_pona()`.

        Message must already be preprocessed, normally done in
        `self.are_toki_pona(message)`.
        """
        scorecards: List[Scorecard] = list()
        for sentence in self.sent_tokenize(message):
            result = self._is_toki_pona(sentence)
            scorecards.append(result)
        scorecards = self.score_sentences(scorecards)
        return scorecards

    def are_toki_pona(self, message: str) -> List[bool]:
        """Splits a statement into sentences, then determines if each is or is not Toki Pona.
        NOTE: You will need to decide how to score the result. Examples:

        ```
        def all_must_pass(message: str) -> bool:
            return all(ILO.are_toki_pona(message))

        def portion_must_pass(message: str, score: Number = 0.8) -> bool:
            results = ILO.are_toki_pona(message)
            sent_count = len(results)
            passing = results.count(True)
            return (passing / sent_count) >= score
        ```
        """
        message = self.preprocess(message)
        scorecards = self._are_toki_pona(message)
        return [card["score"] >= self.__passing_score for card in scorecards]
